#!/bin/bash
# (c) 2014-2016 Freifunk Hochstift <kontakt@hochstift.freifunk.net>
#
# This script builds the firmware by the environment variables given, the
# first two being mandatory:
#
# BASE        = Gluon Version (tag or commit, i.e. v2014.4)
# BRANCH      = Firmware Branch (stable/testing/experimental)
# VERSION     = the version tag (can only be empty if BRANCH=experimental)
#
# optional:
# AUTOUPDATER = force Autoupdater Branch (stable/testing/experimental/off)
# BROKEN      = 0 (default) or 1, build the untested hardware model firmwares, too 
# BUILD_TS    = build timestamp (format: %Y-%m-%d %H:%M:%S)
# DIRCLEAN    = 1 perform "make dirclean" before build (BRANCH==stable/testing) or 0 perform "make clean" (BRANCH==experimental)
# MAKEJOBS    = number of compiler processes running in parallel (default: number of CPUs/Cores)
# PRIORITY    = determines the number of day a rollout phase should last at most
# SITE_ID     = specific site repository commit-id (leave blank to use HEAD)
# SITE_REPO_FETCH_METHOD = http, everything except "git" will use the HTTP method for fetchting site repo
# TARGETS     = a space separated list of target platforms (if unset, all platforms will be build)
# VERBOSE     = 0 (default) or 1, call the make commands with 'V=s' to see actual errors better


### includes
. functions.sh

### static variables
MY_DIR=$(dirname $0)
MY_DIR=$(readlink -f "${MY_DIR}")
CODE_DIR="${MY_DIR}/src"
SITE_DIR="${CODE_DIR}/site"
PATCH_DIR="${SIZE_DIR}/patches"
OUTPUT_DIR="${MY_DIR}/output"
IMAGE_DIR="${CODE_DIR}/output/images"
MODULE_DIR="${CODE_DIR}/output/modules"
VERSIONS_INFO_DIR="${MY_DIR}/versions"

BUILD_INFO_FILENAME="build-info.txt"
SITE_REPO_URL="git.c3pb.de/freifunk-pb/site-ffho.git"
LANG=C

pushd ${MY_DIR} > /dev/null

### ERROR handling
[ -n "${BASE}" ] || abort "Please specify BASE environment variable (Gluon, i.e. 'v2014.3' or commit-id)."
[ -n "${BRANCH}" ] || abort "Please specify BRANCH environment variable."
[ "${BRANCH}" == "experimental" -o "${BASE}" != "HEAD" ] || abort "HEAD is not an allowed BASE-identifier for non-experimental builds. Either use a tagged commit or the commit-SHA itself."
[ -n "${VERSION}" -o "${BRANCH}" == "experimental" ] || abort "Please specify VERSION environment variable (not necessary for experimental branch)."
[ "${BRANCH}" == "experimental" -o ! -r "${VERSIONS_INFO_DIR}/${VERSION}" ] || abort "There exists a version file for '${VERSION}' ... you are trying to do something really stupid, aren't you?"

### set reasonable defaults for unset environment variables
[ -n "${AUTOUPDATER}" ] || AUTOUPDATER=${BRANCH}
# ToDo: [ "${BASE}" == "master" ] && BASE="origin/master"
[ -n "${BROKEN}" ] || BROKEN=0
[ -n "${BUILD_TS}" ] || BUILD_TS=$(date +"%Y-%m-%d %H:%M:%S")

if [ -z "${DIRCLEAN}" ]; then
	if [ "${BRANCH}" == "experimental" ]; then
		DIRCLEAN=0
	else
		DIRCLEAN=1
	fi
fi

[ -n "${MAKEJOBS}" ] || MAKEJOBS=$(grep -c "^processor" /proc/cpuinfo)
[ -n "${PRIORITY}" ] || PRIORITY=0
[ -n "${SITE_REPO_FETCH_METHOD}" ] || SITE_REPO_FETCH_METHOD="http"
[ -n "${TARGETS}" ] || TARGETS=$(BROKEN=${BROKEN} make list-targets | sed -e 's/^[[:space:]]*//' -e 's/[[:space:]]*$//')

if [ "${BRANCH}" == "experimental" -a -z "${VERSION}" ] ; then
	default_release_pattern=$( awk -F" := " '/^DEFAULT_GLUON_RELEASE/ { gsub("shell ", "", $2); print $2; }' ${SITE_DIR}/site.mk )
	VERSION=$(eval echo ${default_release_pattern})
	info "EXPERIMENTAL FIRMWARE: using version tag '$VERSION'"
fi

if [ "${SITE_REPO_FETCH_METHOD}" != "git" ]; then
	SITE_REPO_URL="https://${SITE_REPO_URL}"
else
	SITE_REPO_URL="git@${SITE_REPO_URL}"
fi

MAKE_PARAM=""
[ "$VERBOSE" -eq "1" ] && MAKE_PARAM="${MAKE_PARAM} V=s"

### INIT /src IF NECESSARY
if [ ! -d "${CODE_DIR}" ]; then
	info "Code directory does not exist yet - fetching Gluon ..."
	git clone https://github.com/freifunk-gluon/gluon.git "${CODE_DIR}"
	[ "$?" -eq "0" ] || abort "Failed to fetch Gluon repository."
fi

### INIT /src/site IF NECESSARY
if [ ! -d "${SITE_DIR}" ]; then
	info "Site repository does not exist, fetching it ..."
	git clone "${SITE_REPO_URL}" "${SITE_DIR}"
	[ "$?" -eq "0" ] || abort "Failed to fetch SITE repository."
fi

pushd ${CODE_DIR}

### CHECKOUT GLUON
progress "Checking out GLUON '$BASE' ..."
# TODO: check if gluon got modified and bail out if necessary
git fetch
git checkout -q ${BASE}
[ "$?" -eq "0" ] || abort "Failed to checkout '$BASE' gluon base version, mimimi." >&2
GLUON_COMMIT=$(git rev-list --max-count=1 HEAD)


### CHECKOUT SITE REPO
progress "Checking out SITE REPO ..."
pushd ${SITE_DIR}
# TODO: check if site got modified locally and bail out if necessary
if [ -z "${SITE_ID}" ]; then
	# no specific site given - get the most current one
	git checkout -q ${BRANCH} ; git pull
	[ "$?" -eq "0" ] || abort "Failed to get newest '${BRANCH}' in site repository, mimimi."
else
	# fetch site repo updates
	git fetch || true
	# commit given - use this one
	git checkout -q ${SITE_ID}
	[ "$?" -eq "0" ] || abort "Failed to checkout requested site commit '${SITE_ID}', mimimi."
fi
SITE_COMMIT=$(git rev-list --max-count=1 HEAD)
popd #${SITE_DIR}

### APPLY PATCHES TO GLUON
progress "Applying Patches ..."
git checkout -B patching base
if [ "$(echo "${PATCH_DIR}"/*.patch)" ]; then
	git am ${PATCH_DIR}/*.patch || (
		git am --abort
		git checkout patched
		git branch -D patching
		false
	)
	[ "$?" -eq "0" ] || abort "Failed to apply patches, mimimi."
fi
git branch -M patched


### DIRCLEAN
if [ -d "./build/" -a "${DIRCLEAN}" -eq "1" ]; then
	progress "Cleaning your build environment ..."
	make dirclean
fi

### PREPARE
progress "Preparing the build environment (make update) ..."
make update
[ "$?" -eq "0" ] || abort "Failed to update the build environment, mimimi."


# we are now ready to produce the firmware images, so let's "save" our state
build_info_path="${OUTPUT_DIR}/${BRANCH}/${BUILD_INFO_FILENAME}"
progress "Saving build information to: ${build_info_path}"
[ -n "${build_info_path}" -a -f "${build_info_path}" ] && rm -f ${build_info_path}
mkdir -p $(dirname ${build_info_path})
[ "$?" -eq "0" ] || abort "Unable to create output directory: $(dirname ${build_info_path})"
touch $(dirname ${build_info_path})
[ "$?" -eq "0" ] || abort "Cannot create build information file: ${build_info_path}"
echo "VERSION=${VERSION}" >> ${build_info_path}
echo "GLUON=${GLUON_COMMIT} # ${BASE}" >> ${build_info_path}
echo "BRANCH=${BRANCH}" >> ${build_info_path}
echo "SITE=${SITE_COMMIT} # ${VERSION}" >> ${build_info_path}
echo "TARGETS=${TARGETS}" >> ${build_info_path}
echo "TS=${BUILD_TS}" >> ${build_info_path}

### BUILD FIRMWARE
progress "Building the firmware - please stand by!"

for target in ${TARGETS} ; do
	# configure build environment for our current target
	export GLUON_TARGET="${target}"
	gluon_build_env_vars="GLUON_TARGET=\"${target}\" GLUON_RELEASE=\"${VERSION}\" BROKEN=\"${BROKEN}\""
	[ "${AUTOUPDATER}" != "off" ] && gluon_build_env_vars="${gluon_build_env_vars} GLUON_BRANCH=\"${AUTOUPDATER}\""

	# prepare build environment for our current target
	progress "Preparing build environment for target ${target}."
	[ "${DIRCLEAN}" -eq "0" ] && make clean
	make -j ${MAKEJOBS} prepare-target${MAKE_PARAM}
	[ "$?" -eq "0" ] || abort "Unable to build environment for target-platform ${target}."

	# need to have a toolchain for the particular target 
	progress "Building toolchain for target ${target}."
	make -j ${MAKEJOBS} toolchain/install${MAKE_PARAM}
	[ "$?" -eq "0" ] || abort "Unable to build toolchain for target-platform ${target}."

	# now we can start building the images for the target platform
	progress "Building FFHO-flavoured Gluon firmware for target ${target}. You'd better go and fetch some c0ffee!"
	eval "${gluon_build_env_vars} faketime \"${BUILD_TS}\" make -j ${MAKEJOBS} prepare${MAKE_PARAM}"
	[ "$?" -eq "0" ] || abort "Failed to build firmware for target-platform ${target}."

	# finally compile the firmware binaries
	progress "Compiling binary firmware images."
	faketime "${BUILD_TS}" make -j ${MAKEJOBS} images${MAKE_PARAM}
	[ "$?" -eq "0" ] || abort "Failed to assemble images for target-platform ${target}."

	# compile the modules
	process "Compiling modules."
	eval "${gluon_build_env_vars} faketime \"${BUILD_TS}\" make -j ${MAKEJOBS} modules${MAKE_PARAM}"
	[ "$?" -eq "0" ] || abort "Failed to build modules for target-platform ${target}."
done

popd #${CODE_DIR}

# compress all binaries into 7z archive
progress "Assembling images.7z ..."
pushd ${IMAGE_DIR}
[ -e "${OUTPUT_DIR}/${BRANCH}/images.7z" ] && rm "${OUTPUT_DIR}/${BRANCH}/images.7z"
7z a -xr!*.manifest "${OUTPUT_DIR}/${BRANCH}/images.7z" ./sysupgrade/* ./factory/*
[ "$?" -eq "0" ] || abort "Failed to assemble images (did you install p7zip-full?)."
popd #${IMAGE_DIR}

# compress modules into 7z archive
progress "Assembling modules.7z ..."
pushd ${MODULE_DIR}
[ -e "${OUTPUT_DIR}/${BRANCH}/modules.7z" ] && rm "${OUTPUT_DIR}/${BRANCH}/modules.7z"
7z a "${OUTPUT_DIR}/${BRANCH}/modules.7z" ./gluon-ffho-${VERSION}
[ "$?" -eq "0" ] || abort "Failed to assemble modules."
popd #${MODULE_DIR}

# generate and copy manifests
progress "Generating and copying manifest ..."
pushd ${CODE_DIR}
faketime "${BUILD_TS}" GLUON_PRIORITY=${PRIORITY} GLUON_BRANCH=${BRANCH} make manifest
[ "$?" -eq "0" ] || abort "Failed to generate the manifest, try running 'make manifest' in '$CODE_DIR' directory manually."
cp "${CODE_DIR}/output/images/sysupgrade/${BRANCH}.manifest" "${OUTPUT_DIR}/${BRANCH}/"
popd #${CODE_DIR}

# Saving a copy of the build info file as reference
progress "Building a greater and brighter firmware finished successfully. Saving build information at: ${VERSIONS_INFO_DIR}/${VERSION}"
cp -p "${build_info_path}" "${VERSIONS_INFO_DIR}/${VERSION}"

# The end. Finally.
success "We're done, go and enjoy your new firmware in ${OUTPUT_DIR}/${BRANCH}!"
popd > /dev/null #${MY_DIR}

